from __future__ import annotations
from typing import Any, Callable, TypeVar, cast, overload
from guayaba.app import app, route_exists
from guayaba.utils import base_path_for_post
from typing_extensions import dataclass_transform
from guayaba.endpoints import generate_get, generate_post, generate_put, generate_delete
from starlette.routing import compile_path
from pydantic.dataclasses import dataclass as _pydantic_dataclass

T = TypeVar("T")


# Make pydantic happy
@dataclass_transform()
@overload
def dataclass(route: str, /, **kwargs: Any) -> Callable[[type[T]], type[T]]: ...
@overload
def dataclass(**kwargs: Any) -> Callable[[type[T]], type[T]]: ...


def dataclass(route: str | None = None, /, **kwargs: Any):
    """
    @dataclass("/things/{id}") registers:
      - GET     /things/{id}
      - POST    /things
      - PUT     /things/{id}
      - DELETE  /things/{id}
    The key field ('id', 'name', etc.) is inferred from the route.
    """
    _, _, param_convertors = compile_path(route) if route else (None, None, None)
    key_field = param_convertors and next(iter(param_convertors.keys()))
    if not key_field or not isinstance(key_field, str):
        raise ValueError("Route must include a path parameter like '/path/{id}' or '/path/{name}'.")
    if not isinstance(route, str):
        raise ValueError("Route must be a string like '/path/{id}' or '/path/{name}'.")
    
    def decorator(cls: type[T], /) -> type[T]:
        dc = _pydantic_dataclass(cls, **kwargs)
        base_path = base_path_for_post(route)

        # --- CRUD endpoints ---
        get = generate_get(key_field)
        post = generate_post(key_field)
        put = generate_put(key_field)
        delete = generate_delete(key_field)

        # --- Register routes ---
        if not route_exists(route, "GET"):
            app.add_api_route(route, get, methods=["GET"], name=f"Get {dc.__name__}", response_model=dc)
        if not route_exists(base_path, "POST"):
            app.add_api_route(base_path, post, methods=["POST"], name=f"Create {dc.__name__}", response_model=dc)
        if not route_exists(route, "PUT"):
            app.add_api_route(route, put, methods=["PUT"], name=f"Update {dc.__name__}", response_model=dc)
        if not route_exists(route, "DELETE"):
            app.add_api_route(route, delete, methods=["DELETE"], name=f"Delete {dc.__name__}", response_model=dc)

        return cast(type[T], dc)

    return decorator




import unittest
import guayaba as guba
from fastapi.testclient import TestClient


@guba.dataclass("/hello/{id}")
class User:
    id: int
    name: str
    tags: list[str]


class TestClassID(unittest.TestCase):
    def setUp(self) -> None:
        app = guba.get_app()
        self.client = TestClient(app)

    def test_get(self):
        # Now, test the GET endpoint
        response = self.client.get("/hello/1")
        print(response)

    def test_post(self):
        # Now, test the POST endpoint
        response = self.client.post("/hello/1", json={"name": "John", "tags": ["tag1", "tag2"]})
        print(response)
    
    def test_put(self):
        # Now, test the PUT endpoint
        response = self.client.put("/hello/1", json={"name": "Jane", "tags": ["tag3"]})
        print(response)

    def test_delete(self):
        # Now, test the DELETE endpoint
        response = self.client.delete("/hello/1")
        print(response)




from typing import Any
from fastapi import HTTPException, status, Body
from guayaba.app import output


def generate_get(key: str):
    """Generates a GET endpoint function for the given key field."""

    def fn(**kwargs):
        """"""
        print(key, kwargs)
        key_value = kwargs[key]

        path = output / f"{key_value}.json"
        # TODO: Read pydantic model from file is under the output /path/<key_value>.json

    fn.__annotations__ = {key: Any, "return": Any}
    return fn


def generate_post(key: str):
    """Generates a POST endpoint function for the given key field."""
    def fn(item = Body(...)):
        """"""
        print(key, item)

        path = output / f"{key_value}.json"
        key_value = getattr(item, key, None)
        if key_value is None:
            raise HTTPException(status.HTTP_422_UNPROCESSABLE_ENTITY, detail=f"Item must include a non-empty '{key}' field.")
        # TODO: Write pydantic model to file is under the output /path/<key_value>.json
        return item
    
    fn.__annotations__ = {"item": Any, "return": Any}
    return fn


def generate_put(key: str):
    """Generates a PUT endpoint function for the given key field."""
    def fn(key_value: Any, item = Body(...)):
        """"""
        path = output / f"{key_value}.json"
        # TODO: Write pydantic model to file is under the output /path/<key_value>.json
        return item
    
    fn.__annotations__ = {key: Any, "item": Any, "return": Any}
    return fn


def generate_delete(key: str):
    """Generates a DELETE endpoint function for the given key field."""
    def fn(**kwargs):
        """"""
        path = output / f"{key_value}.json"
        key_value = kwargs[key]
        # TODO: Delete from file is under the output /path/<key_value>.json
        return {"detail": f"Deleted item with {key} '{key_value}'"}
    
    fn.__annotations__ = {key: Any, "return": Any}
    return fn



import re
import os
import json
from typing import Type, TypeVar
from pydantic import BaseModel
from guayaba.app import output
from dataclasses import asdict, is_dataclass
from pydantic.type_adapter import TypeAdapter


T = TypeVar("T", bound=BaseModel)


def base_path_for_post(path: str) -> str:
    suffix_match = re.search(r"/{\w+}$", path)
    return path[:suffix_match.start()] if suffix_match else path


def save(data: object, path: str) -> None:
    """Save any class instance (preferably Pydantic model) or a string to a file."""
    outpath = output / path
    os.makedirs(outpath.parent, exist_ok=True)
    with open(outpath, "w", encoding="utf-8") as f:
        if isinstance(data, BaseModel):
            f.write(data.model_dump_json(indent=4))
        elif is_dataclass(data):
            json.dump(asdict(data), f, indent=4)
        else:
            f.write(str(data))


def load(path: str, model: Type[T] | None = None) -> T | str | None:
    """
    Load a file. If `model` is provided, parse JSON into that Pydantic model.
    Returns:
        - model instance if `model` provided
        - raw string otherwise
        - None if file not found
    """
    try:
        with open(path, "r", encoding="utf-8") as f:
            content = f.read()
            if model:
                if isinstance(model, type) and issubclass(model, BaseModel):
                    return model.model_validate_json(content)
                if is_dataclass(model):
                    adapter = TypeAdapter(model)
                    return adapter.validate_json(content)
                return TypeAdapter(model).validate_json(content)
            return content
    except FileNotFoundError:
        return None
    
    


from enum import Enum
from typing import Optional
import guayaba as guba


class FuelType(Enum):
    PETROL = "Petrol"
    DIESEL = "Diesel"
    ELECTRIC = "Electric"
    HYBRID = "Hybrid"


class TransmissionType(Enum):
    MANUAL = "Manual"
    AUTOMATIC = "Automatic"


@guba.dataclass("/cars/{name}")
class Car:
    name: str
    maker: str
    model: str
    year: int
    fuel_type: Optional[FuelType]
    transmission: Optional[TransmissionType]
    range: float
    price: int



import json
import shutil
import tempfile
import unittest
import guayaba as guba
from pathlib import Path


@guba.dataclass("/hello/{id}")
class User:
    id: int
    name: str
    tags: list[str]


class TestSave(unittest.TestCase):
    def setUp(self):
        output = tempfile.mkdtemp()
        guba.get_app(output)
        self.output = Path(output)

    def tearDown(self):
        shutil.rmtree(self.output)

    def test_save(self):
        user = User(id=1, name="Alice", tags=["tag1", "tag2"])
        guba.save(user, "user_1.json")

        saved_file = self.output / "user_1.json"

        self.assertTrue(saved_file.exists(), "Expected save() to create the output file")

        with open(saved_file, "r", encoding="utf-8") as f:
            content = json.load(f)

        self.assertEqual(
            content,
            {"id": 1, "name": "Alice", "tags": ["tag1", "tag2"]},
            "Saved JSON content does not match the model data",
        )

    def test_load(self):
        saved_file = self.output / "user_1.json"
        saved_file.write_text(
            json.dumps({"id": 1, "name": "Alice", "tags": ["tag1", "tag2"]}),
            encoding="utf-8",
        )

        loaded_user = guba.load(str(saved_file), User)

        self.assertIsInstance(loaded_user, User)
        self.assertEqual(loaded_user.id, 1)
        self.assertEqual(loaded_user.name, "Alice")
        self.assertEqual(loaded_user.tags, ["tag1", "tag2"])




import os
from pathlib import Path
from fastapi import FastAPI
from fastapi.routing import APIRoute


app = FastAPI()
output = Path("docs")


def get_app(path: str = "docs") -> FastAPI:
    global app, output
    output = Path(path)
    os.makedirs(output, exist_ok=True)
    return app


def route_exists(path: str, method: str | None = None) -> bool:
    for r in app.router.routes:
        if isinstance(r, APIRoute) and r.path == path:
            if method is None or method in r.methods:
                return True
    return False


# TODO: Add a /save endpoint to use gitpython to push changes to a branch ready for PR
@app.get("/save")
async def save():
    return {"message": "Save endpoint not implemented yet."}




from pathlib import Path
from setuptools import setup, find_packages

README = Path(__file__).with_name("README.md").read_text(encoding="utf-8")

setup(
    name="guayaba",
    use_scm_version={
        "version_scheme": "guess-next-dev",
        "local_scheme": "no-local-version",
    },
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    description="A sample Python library called Guayaba",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Sau1707",
    author_email="your.email@example.com",
    url="https://github.com/XLongLink/guayaba",
    python_requires=">=3.12",
    install_requires=[
        "fastapi",
    ],
)